/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// host.c -- coordinates spawning and killing of local servers

#include	<setjmp.h>
#include	"../quakedef.h"
#include	"../game.h"
#include	"../menu.h"
#include	"../vid.h"

extern	camera_t	camera;

MR_parms_t			host_parms;

client_t			*host_client;

jmp_buf 			host_abortserver;

cvar_t				max_fps			= {"max_fps",		"100", true};
cvar_t				langame			= {"langame",		"0"};
cvar_t				developer		= {"developer",		"0"};

double				host_frametime;
double				host_time;
double				realtime;
double				oldrealtime;

int					host_hunklevel;
int					host_framecount;

qboolean			host_initialized;

void	Camera_Init	(void);
void	Gamma_Init	(void);
void	Times_Init	(void);
void	S_Init		(void);
void	S_Shutdown	(void);
void	LoadRecords	(void);

/*
================
Host_EndGame
================
*/
void Host_EndGame (char *message, ...)
{
	va_list		argptr;
	char		string[1024];
	
	va_start (argptr,message);
	_vsnprintf (string, sizeof(string), message, argptr);
	va_end (argptr);
	Con_DPrintf ("Host_EndGame: %s\n",string);
	
	if (sv.active)
		Host_ShutdownServer (false);

	if (cls.state == ca_dedicated)
		Sys_Error ("Host_EndGame: %s\n",string);	// dedicated servers exit
	
	if (cls.demonum != -1)
		CL_NextDemo ();
	else
		CL_Disconnect ();

	longjmp (host_abortserver, 1);
}

/*
==========
Host_Error
==========
*/
void Host_Error (char *error, ...)
{
	va_list				argptr;
	char				string[1024];
	static	qboolean	inerror = false;
	
	if (inerror)
		Sys_Error ("Host_Error: recursively entered");
	inerror = true;
	
	SCR_EndLoadingPlaque ();		// reenable screen updates

	va_start (argptr,error);
	_vsnprintf (string, sizeof(string), error, argptr);
	va_end (argptr);
	Con_Printf ("Host_Error: %s\n",string);
	
	if (sv.active)
		Host_ShutdownServer (false);

	if (cls.state == ca_dedicated)
		Sys_Error ("Host_Error: %s\n",string);	// dedicated servers exit

	CL_Disconnect ();
	cls.demonum = -1;

	inerror = false;

	longjmp (host_abortserver, 1);
}

/*
===================
Host_FindMaxClients
===================
*/
void Host_FindMaxClients (void)
{
	if (COM_CheckParm ("-dedicated"))
		cls.state = ca_dedicated;
	else
		cls.state = ca_disconnected;

	if (COM_CheckParm ("-listen"))
		if (cls.state == ca_dedicated)
			Sys_Error ("Only one of -dedicated or -listen can be specified");

	svs.maxclients	= 8;
	svs.clients		= Hunk_AllocName (svs.maxclients*sizeof(client_t), "clients");
}

/*
==============
Host_InitLocal
==============
*/
void Host_InitLocal (void)
{
	Host_InitCommands ();

	Cvar_RegisterVariable (&max_fps);	
	Cvar_RegisterVariable (&langame);
	Cvar_RegisterVariable (&developer);

	Host_FindMaxClients ();
	
	host_time = 1.0;		// so a think at time 0 won't get called
}

/*
=======================
Host_WriteConfiguration
=======================
*/
void Host_WriteConfiguration (void)
{
	FILE	*f;

	if (host_initialized && !isDedicated)
	{
		f = fopen ("data/config.cfg", "w");
		if (!f)
		{
			Con_Printf ("Couldn't write config.cfg.\n");
			return;
		}
		
		Key_WriteBindings (f);
		Cvar_WriteVariables (f);

		fclose (f);
	}
}

/*
===============
SV_ClientPrintf
===============
*/
void SV_ClientPrintf (char *fmt, ...)
{
	va_list		argptr;
	char		string[1024];
	
	va_start (argptr,fmt);
	_vsnprintf (string, sizeof(string), fmt, argptr);
	va_end (argptr);
	
	MSG_WriteByte (&host_client->message, svc_print);
	MSG_WriteString (&host_client->message, string);
}

/*
==================
SV_BroadcastPrintf
==================
*/
void SV_BroadcastPrintf (char *fmt, ...)
{
	va_list		argptr;
	char		string[1024];
	int			i;
	
	va_start (argptr,fmt);
	_vsnprintf (string, sizeof(string), fmt, argptr);
	va_end (argptr);
	
	for (i=0 ; i<svs.maxclients ; i++)
	{
		if (svs.clients[i].active && svs.clients[i].spawned)
		{
			MSG_WriteByte (&svs.clients[i].message, svc_print);
			MSG_WriteString (&svs.clients[i].message, string);
		}
	}
}

/*
===================
Host_ClientCommands
===================
*/
void Host_ClientCommands (char *fmt, ...)
{
	va_list		argptr;
	char		string[1024];
	
	va_start (argptr,fmt);
	_vsnprintf (string, sizeof(string), fmt, argptr);
	va_end (argptr);
	
	MSG_WriteByte (&host_client->message, svc_stufftext);
	MSG_WriteString (&host_client->message, string);
}

/*
=============
SV_DropClient
=============
*/
void SV_DropClient (qboolean crash)
{
	edict_t	*saveSelf;
	int		i;
	client_t *client;

	if (!crash)
	{
		// send any final messages (don't check for errors)
		if (NET_CanSendMessage (host_client->netconnection))
		{
			MSG_WriteByte (&host_client->message, svc_disconnect);
			NET_SendMessage (host_client->netconnection, &host_client->message);
		}
	
		if (host_client->edict && host_client->spawned)
		{
			saveSelf = miniracer.self;
			miniracer.self = host_client->edict;

			ClientDisconnect (host_client->edict);

			miniracer.self = saveSelf;
		}

		Sys_Printf ("Client %s removed\n",host_client->name);
	}

// break the net connection
	NET_Close (host_client->netconnection);
	host_client->netconnection = NULL;

// free the client (the body stays around)
	host_client->active = false;
	host_client->name[0] = 0;
	net_activeconnections--;

// send notification to all clients
	for (i=0, client = svs.clients ; i<svs.maxclients ; i++, client++)
	{
		if (!client->active)
			continue;
		MSG_WriteByte	(&client->message, svc_updatename);
		MSG_WriteByte	(&client->message, host_client - svs.clients);
		MSG_WriteString	(&client->message, "");
		MSG_WriteByte	(&client->message, svc_updatecolors);
		MSG_WriteByte	(&client->message, host_client - svs.clients);
		MSG_WriteByte	(&client->message, 0);
		MSG_WriteByte	(&client->message, 0);
		MSG_WriteByte	(&client->message, 0);
		MSG_WriteByte	(&client->message, svc_updatetime);
		MSG_WriteByte	(&client->message, host_client - svs.clients);
		MSG_WriteFloat	(&client->message, 3599.999f);
		MSG_WriteFloat	(&client->message, 3599.999f);
		MSG_WriteByte	(&client->message, svc_updatelaps);
		MSG_WriteByte	(&client->message, host_client - svs.clients);
		MSG_WriteByte	(&client->message, 0);
	}
}

/*
===================
Host_ShutdownServer
===================
*/
void Host_ShutdownServer(qboolean crash)
{
	int				i;
	int				count;
	sizebuf_t		buf;
	unsigned char	message[4];
	double			start;

	if (!sv.active)
		return;

	sv.active = false;

	if (cls.state == ca_connected)
		CL_Disconnect ();

// flush any pending messages - like the score!!!
	start = Sys_FloatTime();
	do
	{
		count = 0;
		for (i=0, host_client = svs.clients ; i<svs.maxclients ; i++, host_client++)
		{
			if (host_client->active && host_client->message.cursize)
			{
				if (NET_CanSendMessage (host_client->netconnection))
				{
					NET_SendMessage(host_client->netconnection, &host_client->message);
					SZ_Clear (&host_client->message);
				}
				else
				{
					NET_GetMessage(host_client->netconnection);
					count++;
				}
			}
		}
		if ((Sys_FloatTime() - start) > 3.0)
			break;
	}
	while (count);

// make sure all the clients know we're disconnecting
	buf.data = message;
	buf.maxsize = 4;
	buf.cursize = 0;
	MSG_WriteByte(&buf, svc_disconnect);
	count = NET_SendToAll(&buf, 5);
	if (count)
		Con_Printf("Host_ShutdownServer: NET_SendToAll failed for %u clients\n", count);

	for (i=0, host_client = svs.clients ; i<svs.maxclients ; i++, host_client++)
		if (host_client->active)
			SV_DropClient(crash);

//
// clear structures
//
	MR_memset (&sv,			0, sizeof(sv));
	MR_memset (svs.clients,	0, svs.maxclients*sizeof(client_t));
}

/*
================
Host_ClearMemory
================
*/
void Host_ClearMemory (void)
{
	Con_DPrintf ("Clearing memory\n");
	Mod_ClearAll ();
	if (host_hunklevel)
		Hunk_FreeToLowMark (host_hunklevel);

	cls.signon = 0;
	MR_memset (&sv, 0, sizeof(sv));
	MR_memset (&cl, 0, sizeof(cl));
}

/*
===============
Host_FilterTime
===============
*/
qboolean Host_FilterTime (float time)
{
	realtime += time;

	if (!cls.timedemo && realtime - oldrealtime < (1 / max_fps.fvalue))
		return false;		// framerate is too high

	host_frametime	= realtime - oldrealtime;
	oldrealtime		= realtime;

	// don't allow really long or short frames
	if (host_frametime > 0.1)
		host_frametime = 0.1;
	if (host_frametime < 0.001)
		host_frametime = 0.001;
		
	return true;
}

/*
=======================
Host_GetConsoleCommands
=======================
*/
void Host_GetConsoleCommands (void)
{
	int		loop = 1;
	char	*cmd;

	while (loop)
	{
		cmd = Sys_ConsoleInput ();
		if (!cmd)
			break;
		Cbuf_AddText (cmd);
	}
}

/*
================
Host_ServerFrame
================
*/
void Host_ServerFrame (void)
{
	SV_ClearDatagram ();
	
	SV_CheckForNewClients ();

	SV_RunClients ();
	
	if ((svs.maxclients > 1 || key_dest == key_game))
		SV_Physics ();

	SV_SendClientMessages ();
}

/*
==========
Host_Frame
==========
*/
void Host_Frame (float time)
{
	if (setjmp (host_abortserver))
		return;

	rand ();

	if (!Host_FilterTime (time))
		return;
		
	Sys_SendKeyEvents ();

	Cbuf_Execute ();

	NET_Poll();

	if (sv.active)
		CL_SendCmd ();
	
	Host_GetConsoleCommands ();
	
	if (sv.active)
		Host_ServerFrame ();

	if (!sv.active)
		CL_SendCmd ();

	host_time += host_frametime;

	if (cls.state == ca_connected)
		CL_ReadFromServer ();

	SCR_UpdateScreen ();

	host_framecount++;
}

/*
=========
Host_Init
=========
*/
void Host_Init (MR_parms_t *parms)
{
	FILE	*file;

	host_parms	= *parms;
	com_argc	= parms->argc;
	com_argv	= parms->argv;

	Memory_Init (parms->membase);
	Cbuf_Init ();
	Cmd_Init ();	
	Camera_Init ();
	COM_Init ();
	Host_InitLocal ();
	Key_Init ();
	Con_Init ();	
	Menu_Init ();	
	Mod_Init ();
	NET_Init ();
	SV_Init ();

	LoadRecords ();

	if (cls.state != ca_dedicated)
	{
		Gamma_Init ();
		VID_Init ();
		S_Init ();
		Draw_Init ();
		SCR_Init ();
		R_Init ();
		Times_Init ();
		CL_Init ();
	}

	file = fopen ("data/config.cfg","r");

	if (file)
	{
		fclose (file);
		// load the last saved configuration
		Cbuf_InsertText ("exec config.cfg\n");
	}
	else
	{
		// load the base configuration
		Cbuf_InsertText ("exec default.cfg\n");
	}

	// start demos if not already running a server
//	Cbuf_InsertText ("startdemos demo1\n");

	Hunk_AllocName (0, "-HOST_HUNKLEVEL-");
	host_hunklevel = Hunk_LowMark ();

	host_initialized = true;
	
	Sys_Printf ("========MiniRacer Initialized=========\n");	
}

/*
=============
Host_Shutdown
=============
*/

void	Con_CloseDebugLog(void);
void	Draw_Shutdown (void);
void	SaveRecords (void);

void Host_Shutdown (void)
{
	static qboolean isdown = false;
	
	if (isdown)
	{
		printf ("recursive shutdown\n");
		return;
	}
	isdown = true;

// keep Con_Printf from trying to update the screen
	scr_disabled_for_loading = true;

	Host_WriteConfiguration ();
	SaveRecords ();

	NET_Shutdown  ();
	Draw_Shutdown ();
	S_Shutdown ();

	if (cls.state != ca_dedicated)
	{
		VID_Shutdown();
	}
}
